<?php 
/**
* @author  		Michal Baniowski <baniczek@gmail.com>
* @link  		http://www.michalbaniowski.pl
* 
* @package 		bform
*/
abstract class Bform_Core_Driver {

    /**
     * Data container
     * 
     * @var             array
     * @access          protected
     */
    protected $_data = array();
    
    /**
     * Validators container
     * 
     * @var             array
     */
    public $validators = array();
    
    /**
     * Filters container
     * 
     * @var             array
     */
    public $filters = array();
    
    protected static $_allowed_html_tags = array(
        'id',
        'class',
        'title',
        'disabled',
        'readonly',
        'rows',
        'cols',
        'size',
        'style'
    );
   
    public function __construct(Bform_Form $form, $name, array $info = array()) 
    {
        $this->name = $name;
        $this->form = $form;
        
        if (empty($name)) 
        {
            throw new Bform_Exception('Name of the driver cannot be empty!');
        }
        
        if ( ! isset($info['required']))
        {
            $this->required = TRUE;
        }
        
        foreach ($info as $key => $value)
        {
            $this->$key = $value;
        }
        
        if ($this->required)
        {
            $this->validators[] = new Bform_Validator_Driver_Required($this);
            $this->class .= ' '.Kohana::$config->load('bform.css.driver_required_class');
        }
    }

    /**
     * Filter sended value to driver
     * 
     * @return          Bform_Driver 
     * @access          public
     * @final
     * @see             Bform_Form::validate();
     */
    final public function filter() 
    {
        foreach ($this->filters as $f) 
        {
            $f->update();
            $this->value = $f->filter();
        }
        return $this;
    }

    /**
     * Renders a driver
     * 
     * @return          string
     */
    public function render() 
    {
        try
        {
            $content = View::factory($this->driver_template)
                    ->set('driver', $this)->render();
            
            return $content;
        }
        catch (Exception $e)
        {
            Kohana_Exception::handler($e);
        }
    }

    public function __toString() 
    {
        return $this->render();
    }

    public function __set($name, $value) 
    {
        if (preg_match('#([^a-zA-Z0-9\_])#', $name))
        {
            throw new Bform_Exception('Bad value name!');
        }
        
        if ($name[0] === '_')
        {
            $disable_user_method = FALSE;
            $name = substr($name, 1);
        }
        
        $method_name = "_set_$name";
        
        if (method_exists($this, $method_name) AND ! isset($disable_user_method))
        {
            $value = $this->$method_name($value);
        }
        
        $this->_data[$name] = $value;
    }
    
    public function __get($name)
    {
        if ($name[0] === '_')
        {
            $disable_user_method = TRUE;
            $name = substr($name, 1);
        }
        if ( ! array_key_exists($name, $this->_data))
        {
            $this->_data[$name] = NULL;
        }
        
        $value = $this->_data[$name];
        
        $method_name = "_get_$name";
        
        if (method_exists($this, $method_name) AND ! isset($disable_user_method))
        {
            $value = $this->$method_name($value);
        }
        
        return $value;
    }
    
    /**
     *
     * @return          Bform_Form
     */
    public function form()
    {
        return $this->form;
    }
    
    public function __call($name, $arguments)
    {
        $result = FALSE;
        
        if ( ! $this->form()->plugin_manager()->disabled())
        {
            $result = $this->form()->plugin_manager()->execute($name, $arguments, Bform_Plugin_Manager::METHOD_DRIVER, $this);
        }
        
        if ($result & Bform_Plugin::RESULT_NONE)
        {
            if ( ! method_exists($this, $name))
            {
                throw new Bform_Exception('Method :name doesn\'t exists!', array(':name' => $name));
            }
        }
        elseif ($result & Bform_Plugin::RESULT_DRIVER_VALUE)
        {
            return $this->form()->plugin_manager()->value();
        }
        elseif ($result & Bform_Plugin::RESULT_DRIVER)
        {
            return $this;
        }
        else
        {
            throw new exception('Check API!');
        }
    }
    
    protected function _get_driver_layout($value)
    {
        $template_name = $this->form->template();
        $class_name = strtolower(get_class($this));
        
        $path = "bform.drivers.{$class_name}.views.{$template_name}.layout";
        
        $template = Kohana::$config->load($path);
        if (empty($template))
        {
            $path = "bform.views.defaults.drivers_layouts.{$template_name}";
            $template = Kohana::$config->load($path);
        }
        if (empty($template))
        {
            throw new Bform_Exception("Config path ':path' for driver layout doesn't exists!", array(':path' => $path));
        }
        
        return $template;
    }
    
    protected function _get_driver_template($value)
    {
        $template_name = $this->form->template();
        $class_name = strtolower(get_class($this));
        
        $path = "bform.drivers.{$class_name}.views.{$template_name}.driver";
        
        $template = Kohana::$config->load($path);
        if (empty($path))
        {
            throw new Bform_Exception("Config path ':path' for driver template doesn't exists!", array(':path' => $path));
        }
        return $template;
    }
    
    protected function _set_name($value)
    {
        if (substr($value, - 2, 2) == '[]')
        {
            $this->is_array = TRUE;
        }
        
        $tmp = ($this->is_array ? substr($value, 0, strlen($value) - 2) : $value);
        if (strpos($tmp, '[') !== FALSE AND strpos($tmp, ']') !== FALSE) 
        {
            $this->is_namespaced = TRUE;
        }
        
        $path = preg_replace('#\[.*\]#', '', $value);
        preg_match_all('#\[([^\]]+)\]#', $value, $matches);
        foreach ($matches[1] as $n)
        {
            $path .= '.'.$n;
        }
        
        $this->path = $path;
        
        $friendly_name = Bform_Helper::friendly_name($value);

        $this->id = "bform-{$friendly_name}";
        $this->row_id = "bform-row-{$friendly_name}";
        
        return $value;
    }
    
    protected function _get_label($value)
    {
        $value = __($value);
        return $value;
    }
    
    protected function _get_class($value)
    {
        if ( ! empty($value))
        {
            $array = explode(' ', $value);
        }
        else
        {
            $array = array();
        }
        
        if ($this->has_error)
        {
            $array[] = Kohana::$config->load('bform.css.driver_error_class');
        }
        
        $array = array_unique($array);
        $value = implode(' ', $array);
        
        return $value;
    }

    public function _get_html_attributes()
    {
        $attributes = array();
        
        foreach ($this->_data as $name => $value)
        {
            if (in_array($name, self::$_allowed_html_tags))
            {
                $attributes[$name] = $this->$name;
            }
        }
        
        return $attributes;
    }
}