<?php 
/**
* @author  		Michal Baniowski <baniczek@gmail.com>
* @link  		http://www.michalbaniowski.pl
* 
* @package 		bform
*/
abstract class Bform_Core_Form {
    
    const FORM_ID_DRIVER_NAME       = '__form_id';
    
    const VALUES_NOT_EMPTY          = 0x00000001;
    const VALUES_WITHOUT_FORM_ID    = 0x00000010;
    
    const DRIVERS_ONLY_BUTTONS      = 0x00000001;
    const DRIVERS_NOT_BUTTONS       = 0x00000010;
    
    /**
     * Contains Bform_Validator_Exception objects
     * 
     * @var             array
     * @access          protected
     */
    protected $_errors = array();
    
    /**
     * Html name of clicked buttton
     * 
     * @var             string
     * @access          protected
     */
    protected $_submit_value = NULL;
    
    /**
     * @var             array Form params container
     * @access          protected
     */
    protected $_data = array(
        'class'             => 'bform',         // html class
        'name'              => 'bform',         // html name
        'id'                => '',              // html id
        'action'            => '',              // html action
        'layout'            => '',              // 
        'template'          => 'main',          // template of the form
        'method'            => 'post',          // method of sending form
        'data'              => array(),         // loaded data from $_POST or $_GET
        'form_id'           => '',              // ID (class name) of the form
    );
    
    /**
     * @access          public
     * @var             Bform_Driver_Collection
     */
    public $drivers = NULL;
    
    /**
     * Plugin Manager
     * @var             Bform_Plugin_Manager
     * @access          protected
     */
    protected $_plugin_manager = NULL;
    
    /**
     * Form validators
     * @var             array
     * @access          protected
     */
    protected $_validators = array();

    public function __construct(array $options = array())
    {
        foreach ($options as $name => $option)
        {
            $this->param($name, $option);
        }

        $this->drivers              = new Bform_Driver_Collection;
        $this->_plugin_manager      = new Bform_Plugin_Manager($this);
        
        $this->param('form_id', get_class($this));

        $this->method('post');
        $this->action(FALSE);
    }
    
    /**
     * Getter
     * 
     * @return          Bform_Plugin_Manager
     */
    public function plugin_manager()
    {
        return $this->_plugin_manager;
    }
    
    /**
     * Getter for validation errors
     * 
     * @return          array
     * @see             Bform_Form::$_errors
     * @see             Bform_Form::validate()
     * @access          public
     */
    public function errors()
    {
        return $this->_errors;
    }
    
    /**
     * Check if form is sent
     * 
     * @return          bool
     * @access          public
     */
    public function is_sent() 
    {
        if (count($this->param('data')))
        {
            $form_id = $this->param('form_id');
            
            if ( ! empty($form_id))
            {
                $data = $this->data();
                
                if ( ! empty($data[Bform_Form::FORM_ID_DRIVER_NAME]) AND $data[Bform_Form::FORM_ID_DRIVER_NAME] === $form_id) 
                {
                    return TRUE;
                }
            }
        }
        
        return FALSE;
    }
    
    /**
     * Setter and getter for the form params.
     * 
     * @param           string $name Name of the param.
     * @param           string $value If passed - sets the param, otherwise return param
     * @return          Bform_Form
     * @return          mixed
     * @access          public
     */
    public function param($name, $value = NULL)
    {
        if ( ! array_key_exists($name, $this->_data))
        {
            $this->_data[$name] = NULL;
        }
        if ($value === NULL)
        {
            return $this->_data[$name];
        }
        $this->_data[$name] = $value;
        return $this;
    }
    
    /**
     * Template setter / getter
     * 
     * @param           string $value Template name
     * @return          string
     * @return          Bform_Form
     * @access          public
     */
    public function template($value = NULL)
    {
        return $this->param('template', $value);
    }
    
    /**
     * Form send method setter / getter
     * 
     * @param           string $method 'get' or 'post'
     * @return          string
     * @return          Bform_Form
     * @access          public
     */
    public function method($method = NULL) 
    {
        if ($method === NULL) 
        {
            return $this->param('method');
        }
        
        $method = strtolower($method);
        switch ($method) 
        {
            case 'post':
                $load_from = & $_POST;
                break;

            case 'get':
                $load_from = & $_GET;
                break;

            default:
                throw new Bform_Exception('Only POST and GET methods are available!');
        }
        
        $this->data($load_from);
        
        return $this->param('method', $method);
    }

    /**
     * Form action setter / getter
     * 
     * @param           string $value Action URI or URL. If empty string then action will be current URL
     * @return          string
     * @return          Bform_Form
     * @access          public
     */
    public function action($value = NULL) 
    {
        if ($value === NULL)
        {
            return $this->param('action');
        }
        if (empty($value)) 
        {
            $value = "";
        }
        else
        {
            $value = url::site($value);
        }
        return $this->param('action', $value);
    }

    /**
     * This method has to be extended by your class of form
     * 
     * @param           array $params Params to bind
     * @see             Bform::factory()
     */
    abstract public function create(array $params = array());

    /**
     * Adds a filter object to driver
     * 
     * @param           string $driver_name Name of the driver
     * @param           string $filter_class_name Filter class name
     * @param           string $options Params passed into Filter object
     * @return          Bform_Form 
     */
    public function add_filter($driver_name, $filter_class_name, $options = array())
    {
        if ( ! isset($this->drivers->$driver_name))
        {
            throw new Bform_Exception("Driver ':name' do not exists!", array(':name' => $driver_name));
        }
        
        $driver = $this->drivers->$driver_name;

        if ( ! class_exists($filter_class_name))
        {
            throw new Bform_Exception("Filter ':name' do not exists!", array(':name' => $filter_class_name));
        }

        $driver->filters[] = new $filter_class_name($driver, $options);

        return $this;
    }

    /**
     * Add validator to the driver
     * 
     * @param           string $driver_name Name of the driver
     * @param           string $validator_class_name Validator class name
     * @param           array $options Params passed into Validator object
     * @return          Bform_Form 
     */
    public function add_validator($driver_name, $validator_class_name, array $options = array())
    {
        if ( ! isset($this->drivers->$driver_name))
        {
            throw new Bform_Exception("Driver :name doesn't exists!", array(':name' => $driver_name));
        }
        
        $driver = $this->drivers->$driver_name;

        if ( ! class_exists($validator_class_name)) 
        {
            throw new Bform_Exception("Validator :name doesn't exist!", array(':name' => $validator_class_name));
        }

        $driver->validators[] = new $validator_class_name($driver, $options);

        return $this;
    }
    
    /**
     *
     * @param           string $class_name Validator class name
     * @param           array $options Validator params
     * @return          Bform_Form
     * @throws          Bform_Exception 
     */
    public function add_form_validator($class_name, array $options = array())
    {
        if ( ! class_exists($class_name))
        {
            throw new Bform_Exception("Validator :name doesn't exists!", array(':name' => $class_name));
        }
        
        $validator = new $class_name($this, $options);
        if ( ! ($validator instanceof Bform_Validator_Form))
        {
            throw new Bform_Exception('Check API!');
        }
        $this->_validators[] = $validator;
        
        return $this;
    }

    /**
     * Creates a Form ID element, saves a existing $_GET params
     * 
     * @final
     * @access          private
     */
    private final function _after()
    {
        $form_id = $this->param('form_id');
        if (empty($form_id))
        {
            throw new Bform_Exception('Param :name cannot be empty!', array(':name' => 'form_id'));
        }
        $this->add_input_hidden(Bform_Form::FORM_ID_DRIVER_NAME, $form_id);
        
        
        $action = $this->action();
        if (empty($action))
        {
            $action = Request::$initial->url();
            if (($query = Arr::get($_SERVER, 'QUERY_STRING', NULL)) !== NULL AND ! empty($query))
            {
                $action .= '?'.$query;
            }
            $this->action($action);
        }
        
        $layout = $this->param('layout');
        if (empty($layout))
        {
            $this->param('layout', Kohana::$config->load('bform.views.defaults.form_layouts.'.$this->param('template')));
        }
    }

    /**
     * Setter / getter for form sended values.
     * 
     * @param           array $data Array with sended form values
     * @return          Bform_Form
     * @return          array
     */
    public function data(array $data = NULL) 
    {
        return $this->param('data', $data);
    }

    public function __toString()
    {
        return $this->render();
    }

    /**
     * Renders a form.
     * 
     * @return          string
     */
    public function render()
    {
        try 
        {
            $this->_after();

            return View::factory($this->param('layout'))
                    ->set('form', $this)
                    ->render();

        }
        catch (Exception $e) 
        {
            Kohana_Exception::handler($e);
        }
    }
    
    public function __call($function, $values)
    {
        $result = FALSE;
        
        if ( ! $this->plugin_manager()->disabled())
        {
            $result = $this->plugin_manager()->execute($function, $values, Bform_Plugin_Manager::METHOD_FORM);
        }
        
        if ($result & Bform_Plugin::RESULT_FORM_VALUE)
        {
            return $this->plugin_manager()->value();
        }
        elseif ($result & Bform_Plugin::RESULT_FORM OR $result & Bform_Plugin::RESULT_PLUGIN)
        {
            return $this;
        }
        
        if (substr($function, 0, 4) == 'add_')
        {
            $type = substr($function, 4);
            $name = $values[0];

            $driver_class_name = 'Bform_Driver_'.ucfirst($type);

            if ( ! class_exists($driver_class_name))
            {
                throw new Bform_Exception("Class ':name' doesn't exists!", array(':name' => $driver_class_name));
            }

            array_unshift($values, $this);

            $driver = call_user_func_array(array($driver_class_name, 'factory'), $values);

            $this->drivers->{$driver->name} = $driver;

            $this->_update($driver);
            
            return $this;
        }
        
        if ( ! method_exists($this, $function))
        {
            throw new Bform_Exception("Method ':name' doesnt exists!", array(':name' => $function));
        }

        return $this;
    }
    
    /**
     * Alias / Helper for creating button
     * 
     * @param           string $value Html Input value
     * @param           array $info Params
     * @return          Bform_Form 
     */
    public function add_input_submit($value, array $info = array())
    {
        $name = 'submit';
        $this->add_button($name, $value, $info);
        return $this;
    }

    /**
     * Get all values from the form.
     * 
     * @param           integer $get Describes which values should be returned
     * @return          array
     */
    public function values($get = NULL)
    {
        if ($get === NULL)
        {
            $get = (~self::VALUES_NOT_EMPTY) & self::VALUES_WITHOUT_FORM_ID;
        }
        
        $_get_not_empty = ($get & self::VALUES_NOT_EMPTY) ? TRUE : FALSE;
        $_get_without_form_id = ($get & self::VALUES_WITHOUT_FORM_ID) ? TRUE : FALSE;
        
        $_elements = $this->drivers->get_all();

        $values = array();

        foreach ($_elements as $e) 
        {
            if ( ! ($e instanceof Bform_Driver_Html) AND ! ($e instanceof Bform_Driver_Input_File)) 
            {
                if ($_get_without_form_id === TRUE AND $e instanceof Bform_Driver_Input_Hidden AND $e->name == Bform_Form::FORM_ID_DRIVER_NAME)
                {
                    continue;
                }
                
                $e->filter();
                $value = $e->value;
                
                if ($_get_not_empty AND empty($value))
                {
                    continue;
                }
                
                Arr::set_path($values, $e->path, $value, '.');
            }
        }
        
        return $values;
    }

    /**
     * Returns Input Submit value
     * 
     * @return          string
     */
    public function submit_value() 
    {
        return $this->_submit_value;
    }

    /**
     *
     * @param           bool $not_empty
     * @return          array 
     */
    public function files($not_empty = TRUE)
    {
        $drivers = $this->drivers->get_all();
        $files = array();

        foreach ($drivers as $d) 
        {
            if ($d instanceof Bform_Driver_Input_file) 
            {
                if ( ! upload::valid($d->value))
                {
                    continue;
                }
                if ($not_empty AND ! upload::not_empty($d->value))
                {
                    continue;
                }
                $files[$d->name] = $d->value;
            }
        }

        return $files;
    }

    /**
     * Binds saved values to drivers
     * 
     * @param           mixed $data
     * @return          Bform_Form 
     */
    public function bind($data) 
    {
        if (empty($data))
        {
            return $this;
        }
        
        $drivers = $this->drivers->get_all();
        
        foreach ($drivers as $d)
        {
            $path = $d->path;
            $value = Arr::path($data, $path, NULL, '.');
            $d->value = $value;
        }
        return $this;
    }

    /**
     * Validates a form
     * 
     * @return          bool
     */
    public function validate() 
    {
        if ( ! $this->is_sent()) 
        {
            return FALSE;
        }

        $this->_populate_form();

        $_elements = $this->drivers->get_all();

        foreach ($_elements as $element) 
        {
            $element->filter();
            foreach ($element->validators as $v)
            {
                $v->update();
            }
        }

        $errors = array();
        $validated = TRUE;

        foreach ($_elements as $driver_name => $element) 
        {   
            foreach ($element->validators as $v) 
            {   
                try
                {
                    $v->validate();
                }
                catch (Bform_Validator_Exception_Driver $e)
                {
                    $validated = FALSE;
                    
                    $errors[] = $e;
                    
                    if ($e->stop())
                    {
                        $stopped = TRUE;
                        break 2;
                    }
                }
            }
        }
        
        if ( ! isset($stopped))
        {
            foreach ($this->_validators as $v)
            {
                try
                {
                    $v->validate();
                }
                catch (Bform_Validator_Exception_Form $e)
                {
                    $validated = FALSE;
                    
                    $errors[] = $e;
                    
                    if ($e->stop())
                    {
                        $stopped = TRUE;
                        break;
                    }
                }
            }
        }

        $this->_errors = $errors;
        
        call_user_func(array('Bform', 'on_submit'), $this, $validated);

        return $validated;
    }

    /**
     * Fill form with sended data.
     */
    private final function _populate_form()
    {
        if ($this->method() === 'get' AND (bool) count($this->files(FALSE)) === TRUE) 
        {
            throw new Bform_Exception('You cannot send files with GET method!');
        }

        $drivers = $this->drivers->get_all();
        $data = Arr::merge($this->data(), $_FILES);
        
        foreach ($drivers as $d)
        {
            $path = $d->path;
            $value = Arr::path($data, $path, NULL, '.');
            $d->value = $value;
        }
        
        unset($drivers);
        
        $buttons = $this->drivers->get_all(self::DRIVERS_ONLY_BUTTONS);
        foreach ($buttons as $b)
        {
            if (Arr::path($data, $b->path, NULL, '.') !== NULL)
            {
                $this->_submit_value = $b->name;
                break;
            }
        }
    }
    
    private final function _update(Bform_Driver $driver)
    {
        $this->plugin_manager()->update($driver);
    }
    
    public function __set($name, Bform_Driver $value) 
    {
        $this->drivers->$name = $value;
    }

    public function __get($name) 
    {
        if (isset($this->drivers->$name))
        {
            return $this->drivers->$name;
        }
        else 
        {
            throw new Bform_Exception("Element :name doesn't exists!", array(':name' => $name));
        }
    }
}